#!/usr/bin/env python
# -*- coding: utf8 -*-

import hashlib

class Sed:

	def __init__(self):
		self._list = {}

	def add(self, search, replace):
		self._list[search] = replace

	def keys(self):
		t = []
		for k,v in self._list.items():
			t.append(k)
		return t
	
	def get(self, key):
		return self._list[key]

	def getSignature(self):
		sig = ""
		for k,v in self._list.items():
			sig += "%s" % hashlib.sha256("%s%s" % (k,v)).hexdigest()

		return hashlib.sha256(sig).hexdigest()


#!/usr/bin/env python
# -*- coding: utf8 -*-

import collections
import os, stat,re
import functools
import operator

def quacks_like_dict(object):
    """Check if object is dict-like"""
    return isinstance(object, collections.Mapping)
    
def deepmerge(a, b):
    """Merge two deep dicts non-destructively
    
    Uses a stack to avoid maximum recursion depth exceptions
    
    >>> a = {'a': 1, 'b': {1: 1, 2: 2}, 'd': 6}
    >>> b = {'c': 3, 'b': {2: 7}, 'd': {'z': [1, 2, 3]}}
    >>> c = merge(a, b)
    >>> from pprint import pprint; pprint(c)
    {'a': 1, 'b': {1: 1, 2: 7}, 'c': 3, 'd': {'z': [1, 2, 3]}}
    """
    assert quacks_like_dict(a), quacks_like_dict(b)
    dst = a.copy()
    
    stack = [(dst, b)]
    while stack:
        current_dst, current_src = stack.pop()
        for key in current_src:
            if key not in current_dst:
                current_dst[key] = current_src[key]
            else:
                if quacks_like_dict(current_src[key]) and quacks_like_dict(current_dst[key]) :
                    stack.append((current_dst[key], current_src[key]))
                else:
                    current_dst[key] = current_src[key]
    return dst




_rechmod = re.compile(r"(?P<who>[uoga]?)(?P<op>[+\-=])(?P<value>[ugo]|[rwx]*)")
_stat_prefix = dict(u = "USR", g = "GRP", o = "OTH")

def octalmode(location, symbolic):
    """chmod(location, description) --> None
    Change the access permissions of file, using a symbolic description
    of the mode, similar to the format of the shell command chmod.
    The format of description is
        * an optional letter in o, g, u, a (no letter means a)
        * an operator in +, -, =
        * a sequence of letters in r, w, x, or a single letter in o, g, u
    Example:
        chmod(myfile, "u+x")    # make the file executable for it's owner.
        chmod(myfile, "o-rwx")  # remove all permissions for all users not in the group. 
    See also the man page of chmod.
    """

    mo = _rechmod.match(symbolic)
    who, op, value = mo.group("who"), mo.group("op"), mo.group("value")
    if not who:
        who = "a"
    mode = os.stat(location)[stat.ST_MODE]
    if value in ("o", "g", "u"):
        mask = _ors((_stat_bit(who, z) for z in "rwx" if (mode & _stat_bit(value, z))))
    else:
        mask = _ors((_stat_bit(who, z) for z in value))
    if op == "=":
        mode &= ~ _ors((_stat_bit(who, z) for z in  "rwx"))
    mode = (mode & ~mask) if (op == "-") else (mode | mask)
    
    return mode


# Helper functions

def _stat_bit(who, letter):
    if who == "a":
        return _stat_bit("o", letter) | _stat_bit("g", letter) | _stat_bit("u", letter)
    return getattr(stat, "S_I%s%s" % (letter.upper(), _stat_prefix[who]))

def _ors(sequence, initial = 0):
    return functools.reduce(operator.__or__, sequence, initial)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import types
import logging
from puke.Error import *
from puke.Console import *
from pydoc import help
from inspect import getcallargs

__tasks__ = {}


def addTask(task):
    console.debug("Registering task: %s" % task.name)
    __tasks__[task.name] = task


def hasDefault():
    if 'default' not in __tasks__:
        return False
    
    return True
    
def executeTask(name, *args):
    if name in __tasks__:
        console.header("-------------------------------------\n * Executing task: %s (%s) \n -------------------------------------" % (name, __tasks__[name].desc), 1)
        __tasks__[name](*args)
    else:
        raise PukeError("No such task: %s" % name)

def printHelp(name = ""):
    if name in __tasks__:
        console.header("-------------------------------------\n * Help %s (%s) \n -------------------------------------" % (name, __tasks__[name].desc), 1)
        help(__tasks__[name].getFunc())
    else:
        raise PukeError("No such task: %s" % name)
    
        
def printTasks():
    for name in __tasks__:
        obj = __tasks__[name]
        if obj.desc:
            logging.info("%s: %s" % (name, obj.desc))
        else:
            logging.info("%s" % name)



class Task:
    __doc__ = ""
    
    def __init__(self, func, desc=""):
        name = func.__name__
        self.__func = func
        
        self.name = name
        self.desc = desc
        self.fullname = "%s.%s" % (func.__module__, name)
        
        try:
            self.__doc__ = func.__doc__
        except AttributeError:
            pass
            
        addTask(self)
        

    def __call__(self, *args, **kw):
        #try:
        try:
             getcallargs(self.__func, *args)
        except TypeError as e:
           printHelp(self.name)
           console.fail("%s" % e)
           raise
       

        retval = self.__func(*args)
        return retval


    def __repr__(self):
        return "Task: " + self.name

    def getFunc(self):
        return self.__func



def task(func):
    """ Specifies that this function is a task. """
    
    if isinstance(func, Task):
        return func

    elif isinstance(func, types.FunctionType):
        return Task(func)
    
    else:
        # Used for descriptions
        def wrapper(finalfunc):
            return Task(finalfunc, func)
            
        return wrapper

    


#!/usr/bin/env python
# -*- coding: utf8 -*-

import urllib2
import hashlib

from puke.Console import *
from puke.FileSystem import *

import urlparse, StringIO, os, ftplib

class Cache:

	@staticmethod
	def fetchHttp(url):
		id = hashlib.sha256(url).hexdigest()

		if Cache.check(id):
			return Cache.getPath(id)
		
		infos = urlparse.urlparse(url)
		try:
			console.debug('Cache#Fetching remote file %s' % url)

			if infos.scheme == 'ftp':
				buffer = StringIO.StringIO()

				ftp = ftplib.FTP(infos.netloc)
				ftp.login('anonymous')
				ftp.cwd(os.path.dirname(infos.path))
				ftp.retrbinary("RETR " + os.path.basename(infos.path), buffer.write)
				payload = buffer.getvalue()
				buffer.close()
			else:
				handler = urllib2.build_opener()
				payload = handler.open(url, None, 60).read()
		except Exception as error:
			console.fail('HTTP fail %s (%s)' % (url, error))
			return False

		return Cache.write(id, payload, url)


	@staticmethod
	def write(id, payload, info = None):

		writefile(join(".pukecache", "%s" % id), payload)

		if info:
			writefile(join(".pukecache", "%s.meta" % id), info)

		return Cache.getPath(id)
		
	@staticmethod
	def read(id):
		if not Cache.check(id):
			return None
		
		return readfile(join(".pukecache", "%s" % id))

	@staticmethod
	def check(id):
		return exists(join(".pukecache", "%s" % id))

	@staticmethod
	def getPath(id):
		return join(".pukecache", "%s" % id)

	@staticmethod
	def getInfo(id):
		if not exists(join(".pukecache","%s.meta" % id)):
			return ""
		
		return readfile(join(".pukecache", "%s.meta" % id))


	@staticmethod
	def clean():
		try:
			remove(".pukecache")
			return True
		except:
			return False


#!/usr/bin/env python
# -*- coding: utf8 -*-

import tarfile
import zipfile
import datetime, time

class Compress:

	@staticmethod
	def open(file, mode):

		ext = file.split('.')[-1]
		if ext == "zip":
			i =  CompressZip()
		elif ext == "bz2":
			mode += ":bz2"
			i = CompressTar()
		else:
			mode += ":gz"
			i = CompressTar()
		
		i.open(file, mode)

		return i
	
	@staticmethod
	def check(file):
		if CompressTar.check(file):
			return True

		if CompressZip.check(file):
			return True
		
		return False




class CompressInterface():
	
	def __init__(self):
		'''Init'''
		raise NotImplemented()

	def open(self,file, mode):
		'''init compress instance'''
		raise NotImplemented()

	def add (self, file):
		''' add file'''
		raise NotImplemented()

	def close(self):
		'''close it'''
		raise NotImplemented()

	@staticmethod
	def check(file):
		'''validate pack format'''
		raise NotImplemented()
		

	def extract(self, file):
		'''uncompress file'''
		raise NotImplemented()

	def __iter__(self):
		raise NotImplemented()

class CompressInfo():

	mode = 0
	mtime = 0

class CompressTar(CompressInterface):
	def __init__(self):
		self.__instance = None

	def open(self, file, mode):
		self.__instance = tarfile.open(file, mode)
	
	def add(self, file):
		self.__instance.add(file)

	@staticmethod
	def check(file):
		return tarfile.is_tarfile(file)

	def extract(self, file):
		res = self.__instance.extractfile(file)

		if not res:
			return (None, None)

		infos = CompressInfo()
		tarinfo = self.__instance.getmember(file)

		infos.mode = int("%o" % tarinfo.mode)
		infos.mtime = tarinfo.mtime
		
		return (res.read(), infos)

	def __iter__(self):
		return iter(self.__instance.getnames())

	def close(self):
		self.__instance.close()





class CompressZip(CompressInterface):
	
	def __init__(self):
		self.__instance = None

	def open(self, file, mode):
		self.__instance = zipfile.ZipFile(file, mode)
	
	def add(self, file):
		self.__instance.write(file)

	@staticmethod
	def check(file):
		return zipfile.is_zipfile(file)

	def extract(self, file):

		infos = CompressInfo()

		infos.mode = int( "%o" % ((self.__instance.getinfo(file).external_attr >> 16L) & 0777))
		infos.mtime = time.mktime(datetime.datetime(*self.__instance.getinfo(file).date_time[0:6]).timetuple())
		return (self.__instance.read(file), infos)

	def __iter__(self):
		for item in self.__instance.namelist():
			if item.endswith('/'):
				continue

			yield item


	def close(self):
		self.__instance.close()





#!/usr/bin/env python
# -*- coding: utf8 -*-

import platform

if platform.system() == 'Windows':
    from compat import *

from puke.Error import *
from puke.Task  import *
from puke.Tools  import *
from puke.ToolsExec  import *
from puke.FileList import *
from puke.Sed import *
from puke.Console import *
from puke.Env import *
from puke.Cache import *
from puke.Require import *
from puke.Yak import *
from puke.VirtualEnv import *
from puke.Std import *
import puke.System
import puke.FileSystem
import puke.Utils


VERSION = 0.1
__all__ = [
            "main", "VERSION", "Error", "FileList", "Sed", "Std", "Env", "Require", "Yak", "VirtualEnv", "System", "FileSystem", "Utils",
            "combine", "sh", "minify", "jslint", "jsdoc", "patch", "prompt", "deepcopy", "stats", "pack", "unpack", "hsizeof", "console"
         ]


import sys, logging, os, traceback
import pkg_resources
from optparse import OptionParser

from colorama import *



try:
    sys.path.insert(1, os.getcwd())
except:
    pass

def run():
    """ Main routine which should be called on startup """

    #
    # Parse options
    #

    parser = OptionParser()
    parser.add_option("-c", "--clear", action="store_true", dest="clearcache", help="Spring time, clean all the vomit")
    parser.add_option("-q", "--quiet", action="store_false", dest="verbose", help="don't print status messages to stdout")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", help="print more detailed status messages to stdout")
    parser.add_option("-t", "--tasks",action="store_true",  dest="list_tasks", help="list tasks")
    parser.add_option("-l", "--log", dest="logfile", help="Write debug messages to given logfile")
    parser.add_option("-f", "--file", dest="file", help="Use the given build script")
    parser.add_option("-p", "--patch",action="store_true",  dest="patch", help="Patch closure")
    parser.add_option("-i", "--info",action="store_true",  dest="info", help="puke task --info show task informations")

    (options, args) = parser.parse_args()


    #
    # Configure logging
    # 

    if options.logfile:
        logging.basicConfig(filename=options.logfile, level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s")
        logging.getLogger().setLevel(logging.DEBUG)
    elif not logging.root.handlers:
        if options.verbose is True:
            logging.getLogger().setLevel(logging.DEBUG)
        elif options.verbose is False:
            logging.getLogger().setLevel(logging.WARN)
        else:
            logging.getLogger().setLevel(logging.INFO)


    
    


    
    # Define a Handler which writes INFO messages or higher to the sys.stderr
    consoleCfg = logging.StreamHandler()

    if options.verbose is True:
        consoleCfg.setLevel(logging.DEBUG)
    elif options.verbose is False:
        consoleCfg.setLevel(logging.WARN)
    else:
        consoleCfg.setLevel(logging.INFO)
    
    if os.environ.get("NOCOLOR"):
        consoleCfg.setFormatter(logging.Formatter( ' %(message)s' , '%H:%M:%S'))
    else:
        consoleCfg.setFormatter(logging.Formatter( ' %(message)s' + Style.RESET_ALL, '%H:%M:%S'))

    logging.getLogger().addHandler(consoleCfg)



     #Patch closure

    try:
        closure = pkg_resources.get_distribution('closure_linter').location
        closure_lock = os.path.join(closure, 'closure_linter', 'puke.lock')
        
        if options.patch or not os.path.isfile(closure_lock):
            closure = os.path.join(closure, 'closure_linter', 'ecmalintrules.py')
            
            try:
                handle = source = destination = None
                import shutil
                shutil.move( closure, closure+"~" )
                destination= open( closure, "w" )
                source= open( closure+"~", "r" )
                
                content = source.read()
                content = content.replace('MAX_LINE_LENGTH = 80', 'MAX_LINE_LENGTH = 120')
                destination.write(content)

                source.close()
                destination.close()

                os.remove(closure+"~" )


                handle = file(closure_lock, 'a')
                handle.close

                if options.patch:
                    console.confirm('Patch successful')
                    sys.exit(0)
            except Exception as e:
                console.warn(">>> you should consider running \"sudo puke --patch\"")
                if handle:
                    handle.close()
                if source:
                    source.close()
                if destination:
                    destination.close()
                
                if options.patch:
                    sys.exit(0)


    except Exception as e:
        console.error('Closure linter not found %s' % e)


    #
    # Find and execute build script
    #
    
    pukefiles = ["pukefile", "pukeFile", "pukefile", "pukefile.py", "pukeFile.py", "pukefile.py"]
    
    script = None
    if options.file:
        if os.path.isfile(options.file):
            script = options.file
    else:
        for name in pukefiles:
            if os.path.isfile(name):
                script = name

    if script is None:
        if options.file:
            raise PukeError("No generate file '%s' found!" % options.file)
        else:
            raise PukeError("No generate file found!")
    
    retval = execfile(script)

    #
    # Execute tasks
    #


    if options.list_tasks:
        console.confirm("Please choose from: ")
        printTasks()
        sys.exit(0)

   

    
    if options.clearcache:
        console.header("Spring time, cleaning all the vomit around ...")
        console.log("...")
        if Cache.clean():
            console.confirm("You're good to go !\n")
        else:
            console.confirm("Your room is already tidy, good boy :-) \n")
        sys.exit(0)
    
    try:
        args = args.strip()
    except:
        pass

    if not args:
        if hasDefault():
            executeTask('default')
        else:
            logging.error("No tasks to execute. Please choose from: ")
            printTasks()
            sys.exit(1)

    else:
        name = args.pop(0)

        if options.info:
            printHelp(name.strip())
        else:
            executeTask(name.strip(), *args)
                


def gettraceback(level = 0):
    trace = ""
    exception = ""
    exc_list = traceback.format_exception_only (sys.exc_type, sys.exc_value)

    reverse = -1 - level
    for entry in exc_list:
        exception += entry
    tb_list = traceback.format_tb(sys.exc_info()[2])
    for entry in tb_list[reverse]:
        trace += entry  
    return trace  

def main():
    try:
        run()
    
    except Exception as error:

        console.fail("\n\n :puke: \n PUKE %s \n %s \n" % (error, gettraceback()))

        sys.exit(1)
        
    except KeyboardInterrupt:
        console.warn("\n\n :puke: \nBuild interrupted!\n")
        sys.exit(2)
        
    sys.exit(0)


#!/usr/bin/env python
# -*- coding: utf8 -*-

"""REquire parser

This class simply provide an accessor on a dict
"""


'''generic:
	titi: "dtc"

dev:
	toto: "${BOURSE_PATH}|/toto/coucou"
	titi: "${BOURSE_PATH}|/toto/coucou"
	coin:
		toto: 'cul'




c = Require("mesbourses.yaml")



#c.merge("toto.yaml")
c.makevars('generic')
-> titi, valeur dtc
c.makevars('dev')
-> titi, jgjg

#make_glob(c.get('dev'))

global toto
global titi

toto = c.get('dev')['toto']
titi = c.get('dev')['titi']

coin = [toto: vf,;:]





CONFIG['toto']'''

import os,re
import yaml,json
from puke.Env import *
from puke.Console import *
from puke.Yak import *
from puke.Utils import *



class Require(object):

	__sharedState = {}
	__globalPattern = re.compile('\$\{([^}]+)\}([|])?(.*)')


	def __init__(self, filename):
		self.__dict__ = self.__sharedState
		if not self.__sharedState:
			self.__files = [filename]
			self.__cfg = self.__load(filename)
			self.__makeenvs(self.__cfg)
			

	def merge(self, filename):
		self.__files.append(filename)
		self.__cfg = deepmerge(self.__cfg, self.__load(filename))
		self.__makeenvs(self.__cfg)

	def yak(self, selector):
		if not self.get(selector):
			return False

		for (node, value) in self.get(selector).items():
			existing = Yak.get(node)
			if existing != None and not isinstance(existing, str):
				value = deepmerge(Yak.get(node), value)
							
			Yak.set(node,value)
		
		

	def get(self, key):
		if self.__cfg.has_key(key):
			return self.__cfg[key]

	def set(self, key, value):
		self.__cfg[key] = value

	def __getitem__(self, key):
		return self.get(key)

	def __setitem__(self, key, value):
		self.set(key, value)

	def __contains__(self, key):
		return (key in self.__cfg)

	def __repr__(self):
		return 'Config: %s' % self.__cfg

	def __load(self, filename):
		try:
			stream = None
			stream = file(filename, 'r')

			payload = stream.read()
			stream.close()

			size = payload.strip()

			ext = os.path.splitext(stream.name)[1]

			result = None

			if size == 0:
				result = { }
				return result
			
			if ext in ['.json', '.js']:
				result = json.loads(payload)
			elif ext in ['.yaml', '.yml']:
				result = yaml.load(payload)
						
		except Exception as error :
			raise RequireError("Require load error : %s" % error)
			result = None

		return result
	
	def __makeenvs(self, data ):
		if data == None:
			return False

		if isinstance(data, list):
			dataIter = enumerate(data)
		else:
			dataIter = data.items()
		
		for (node, value) in dataIter:
			if not isinstance(value, (str, int)):
				self.__makeenvs(value)
			elif value.startswith('${'):
				m = self.__globalPattern.match(value)
				if not m:
					continue
				
				(name, isDefault , extValue) = m.groups()
				if isDefault:
					default = extValue
				else:
					default = ''

				value = Env.get(name, default)

				if not isDefault:
					value += extValue

				data[node] = value
				

	def reload(self):
		files = self.__files
		self.__files = []
		self.__cfg = {}
		for filename in files:
			self.include(filename)

class RequireError(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)



#!/usr/bin/env python
# -*- coding: utf8 -*-

class Yak(dict):

	class __metaclass__(type):
		def __iter__(self):
			for attr in dir(Yak):
				if not attr.startswith("__"):
					yield attr
	@staticmethod
	def set(key, value):
		setattr(Yak, key, value)

	@staticmethod
	def get(key, default = None):
		try:
			return getattr(Yak, key)
		except:
			return default



import os, sys

import subprocess
import signal, platform
from puke.Console import *
from puke.Std import *



class Alarm(Exception):
    pass

def alarm_handler(signum, frame):
    raise Alarm


def sh (command, header = None, output = True, timeout = None, std = None):

    isWouinWouin = (platform.system() == 'Windows')

    if isinstance(command, list):
        command = " \n ".join(command)
    



    if header == None:
        header = 'exec "%s" ' % command
    else:
        console.debug(command)
    
    if output and header != False:
        console.header(' - '+header)


    stdErrMarker = "XXXERRORXXX"
    
    result = ""

    if timeout and not isWouinWouin:
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(timeout)  # 5 seconds
    

    try:
        executable = '/bin/bash'
        command = "export PYTHONIOENCODING=utf-8;" + command


        if isWouinWouin:
            executable = None
            command = "bash.exe -c '" + command.replace("'", "\\'") + "'"
        

        cProcess = subprocess.Popen(command, stdout = subprocess.PIPE, shell = True, stderr= subprocess.PIPE, executable=executable, env=os.environ)
        (rStdout, rStderr) = cProcess.communicate()
        code = cProcess.returncode

        if not isWouinWouin:
            signal.alarm(0)

    except Alarm:
        console.debug('taking too long (%s)' % command)
        cProcess.kill()
        rStdout = rStderr = ''
        code = 0
    
    if std and isinstance(std, Std):
        std.set(rStdout, rStderr, code)

    rStdout = "%s\n%s\n%s" % (rStdout if rStdout else '', stdErrMarker if rStderr else '', rStderr if rStderr else '')

    lines = rStdout.split('\n')
    for line in lines:
        if not line:
            continue

        if output == True and stdErrMarker in line:
            console.error("")
            console.error('Std error :')
            line = ""
        

        result += line
        
        if not line.endswith('\n'):
            result += '\n'

        if output == True:
            console.info( '   ' +line)
   
    return result


#!/usr/bin/env python
# -*- coding: utf8 -*-

import os, time, pwd, grp
import shutil, hashlib
from puke.Console import *
import Utils

def makedir(dirname):
    """ Creates missing hierarchy levels for given directory """
    
    if dirname == "":
        return
    
    if exists(dirname) and isfile(dirname):
        raise FileSystemError("%s is a file. Cannot create a directory" % dirname)
        
    if not os.path.exists(dirname):
        try:
            os.makedirs(dirname)
        except Exception as e:
            raise FileSystemError("Error creating %s : %s" % (dirname, e) )

def readfile(file):
    data = None
    fh = None
    try:
        fh = open(file)
        data = fh.read()
        fh.close()
    except Exception as e:
        raise FileSystemError("%s : %s" % (e, file))
    finally:
        if fh:
            fh.close()


    return data

def checksum (file):
    block_size=2**20
    md5 = hashlib.md5()
    data = None
    fh = None
    try:
        fh = open(file, 'rb')
        while True:
            data = fh.read(block_size)
            if not data:
                break
            md5.update(data)

        fh.close()
    except Exception as e:
        raise FileSystemError(e)
    finally:
        if fh:
            fh.close()
    
    
    return md5.hexdigest()

def remove(file):

    if not file or file in ['./', '/', '~/', '~', '.', '..', '../']:
        raise FileSystemError('Invalid path %s' % file)
    
    try:
        if os.path.isfile(file):
            os.remove(file)
        else:
            shutil.rmtree(file)
    except Exception as e:
        raise FileSystemError( "Error removing %s : %s" % (file , e))
    

def copyfile(src, dst, force = False):
    """ Copy src file to dst file. Both should be filenames, not directories. """
    
    if not os.path.isfile(src):
        raise FileSystemError("No such file: %s" % src)
    

    try:
        dst_mtime = os.path.getmtime(dst)
        src_mtime = os.path.getmtime(src)
        
        # Only accecpt equal modification time as equal as copyfile()
        # syncs over the mtime from the source.
        if force and src_mtime <= dst_mtime:
            return False
        
    except OSError:
        # destination file does not exist, so mtime check fails
        pass

    # First test for existance of destination directory
    makedir(os.path.dirname(dst))
    
    # Finally copy file to directory
    try:
        shutil.copy2(src, dst)
    except IOError as ex:
        raise FileSystemError("Could not write file %s: %s" % (dst, ex))
        
    return True
    
def exists(src):
    if not os.path.exists(src):
        return False

    return True
    
def isfile(src):
    if not exists(src):
        raise FileSystemError('isfile error : %s not found' % src)

    if not os.path.isfile(src):
        return False

    return True

def isdir(dirname):
    if not exists(dirname):
        raise FileSystemError('isdir error : %s not found' % dirname)

    if not os.path.isdir(dirname):
        return False

    return True


def writefile(dst, content, mtime = None, binary = False):
    # First test for existance of destination directory
    makedir(os.path.dirname(dst))
    
    # Open file handle and write
    if binary:
        mode = "wb"
    else:
        mode = "w"
    try:
        handle = open(dst, mode=mode)
        handle.write(content)
        handle.close()
    except Exception as e:
        raise FileSystemError("Error writing file %s : %s" % (dst, e))

    if mtime:
        os.utime(dst, (time.time(), mtime))

def join(*args):
    return os.path.join(*args)

def abspath(path):
    return os.path.abspath(path)

def basename(path):
    return os.path.basename(path)

def dirname(path):
    return os.path.dirname(path)

def normpath(path):
    return os.path.normpath(path)

def sep():
    return os.sep

def chown(path, uname = None, gname = None):
    isfile(path)

    if not uname:
        uid = os.stat(path).st_uid

    if not gname:
        gid = os.stat(path).st_gid

    if not isinstance(uname, int) and uname != None:
        uid = getUid(uname)
    
    if not isinstance(gname, int) and gname != None:
        gid = getGid(gname)

    
    if uid == None or gid == None:
        raise FileSystemError('CHOWN %s : Invalid uname or gname (%s->%s:%s->%s)' % (path, uname, uid, gname, gid))

    return os.chown(path, uid, gid)

def chmod(path, mode):
    isfile(path)


    if not isinstance(mode, int):
        mode = Utils.octalmode(path, mode)
    else:
        mode = int("%s" % mode, 8)
    
    return os.chmod(path, mode)

def getUid(name):
    try:
        return pwd.getpwnam(name)[2]
    except:
        return None

def getGid(name):
    try:
        return grp.getgrnam(name)[2]
    except:
        return None



class FileSystemError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


#!/usr/bin/env python
# -*- coding: utf8 -*-

import logging, sys, os, json
from colorama import *

init(autoreset = True)

class console:
	


	@staticmethod
	def log(*messages):
		color = Style.BRIGHT 
		if os.environ.get("NOCOLOR"):
			color = ""

		for m in messages:
			msg = console.stringify(m)
			logging.info(color + msg)
	
	@staticmethod
	def info(*messages):
		for m in messages:
			msg = console.stringify(m)
			logging.info(   msg)

	@staticmethod
	def debug(*messages):
		color = Back.BLUE

		if os.environ.get("NOCOLOR"):
			color = ""
		
		if os.environ.get("NOCOLOR"):
			color = ""

		for m in messages:
			msg = console.stringify(m)
			logging.debug( color +  msg )

	@staticmethod
	def warn(*messages):
		color = Fore.YELLOW + Style.BRIGHT

		if os.environ.get("NOCOLOR"):
			color = ""

		for m in messages:
			msg = console.stringify(m)
			msg = console.pukefactory(msg)
		
			logging.warning( color + msg)
	
	@staticmethod
	def error(*messages):
		color = Back.RED + Style.BRIGHT

		if os.environ.get("NOCOLOR"):
			color = ""

		for m in messages:
			msg = console.stringify(m)
			logging.error( color + msg)

	@staticmethod
	def confirm(*messages):
		color = Fore.GREEN + Style.BRIGHT

		if os.environ.get("NOCOLOR"):
			color = ""
			
		for m in messages:
			msg = console.stringify(m)
			logging.info(color + msg )
	
	@staticmethod
	def header(msg, level = 2):
		msg = console.stringify(msg)
		logging.info("")
		if level == 1:
			color = Fore.MAGENTA
		else:
			color = Fore.CYAN
		
		if os.environ.get("NOCOLOR"):
			logging.info(msg )
		else:
			logging.info(color + Style.BRIGHT + msg )

	@staticmethod
	def fail(msg):
		msg = console.stringify(msg)
		msg = console.pukefactory(msg)
		console.error(" /!\\ BUILD FAIL : " + msg)
		sys.exit(1)

	@staticmethod
	def pukefactory(msg):
		if ':puke:' in msg:
			try:
				f = open(os.path.join(os.path.dirname( __file__ ), 'datas','decoration', 'puke.txt'), 'r')
				msg = msg.replace(':puke:', '\n' + f.read())
				f.close()

				return msg
			except Exception:
				pass
		
		return msg


	@staticmethod
	def stringify(msg):
		if isinstance(msg, str):
			return msg

		return json.JSONEncoder().encode(msg)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os, os.path, shutil, logging, sys, filecmp, stat, re, time


from puke.FileList import *
from puke.Console import *
from scss import Scss
from puke.FileSystem import *
from puke.Compress import *
from puke.Std import *
from puke.ToolsExec import *



from puke.Cache import *
import System





CSS_COMPRESSOR = sys.argv[0] + '.css.compress'
JS_COMPRESSOR = sys.argv[0] + '.js.compress'
JS_RUNNER = sys.argv[0] + '.js.runner'

def combine(in_files, out_file, verbose=False, replace = None):

    in_files = FileList.check(in_files)
         
    builddir = os.path.dirname(out_file)

    makedir(builddir)

    temp_file = os.path.join(builddir, '.temp')

    console.header( "- Combining files :")

    isCSS = False
    combined = ""

    for f in in_files:
        if not __get_ext(f) in ['css', 'scss']:
            continue

        isCSS = True
        combined = "@option compress: no;"
    

    for f in in_files:

        fh = open(f)
        data = fh.read() + '\n'

        if replace:
            data = __replace(data, replace)

        fh.close()
   
        console.info('  + %s ' % (__pretty(f)))

        combined += data
        
    
    if isCSS:
        combined = __parse_scss(combined)

    console.confirm( "  Generating %s" % out_file)
    writefile(out_file, combined)

def minify(in_file, out_file = None, verbose=False):

    System.check_package('java')

    if not isinstance(in_file, str):
        raise Exception("Minify : single file only")

    if not out_file:
        out_file = in_file
    
    in_type = __get_ext(out_file)

    org_size = os.path.getsize(in_file)

    console.header('- Minifying %s (%.2f kB)' % (__pretty(in_file), org_size / 1024.0))

    if in_type == 'js':
        __minify_js(in_file, out_file + '.tmp', verbose)
    else:
        __minify_css(in_file, out_file + '.tmp', verbose)
    
    copyfile(out_file + '.tmp', out_file)
    os.remove(out_file + '.tmp')

    new_size = os.path.getsize(out_file)

    #avoid division by zero
    if not new_size:
        console.fail('Compression fail')

    
    console.info('  ~ Original: %.2f kB' % (org_size / 1024.0))
    console.info('  ~ Compressed: %.2f kB' % (new_size / 1024.0))
    console.confirm('  %s ( Reduction : %.1f%% )' % (out_file, (float(org_size - new_size) / org_size * 100)))


def jslint(files, fix = False, relax = False, fail = True):
    System.check_package('java')

    files = FileList.check(files)

    options = []
    command = ""

    options.append('--jslint_error=optional_type_marker')
    options.append('--jslint_error=blank_lines_at_top_level')
    options.append('--jslint_error=indentation')
    options.append('--custom_jsdoc_tags=homepage,version,ignore,returns,example,function,requires,name,namespace,property,static,constant,default,location,copyright,memberOf,lends,fileOverview')


    if fix == True:
        header = "Fix JS lint"
        command = "fixjsstyle %s %s "  % (  ' '.join(options) , ' '.join(files))
    else:
        header = "JS lint"
        command = "gjslint %s %s "  % (  ' '.join(options) , ' '.join(files))
    
    if relax == True:
        command += ' | grep -v "Line too long"'
    result = sh(command, header = "%s (%s files)" % (header, len(files)) )

    error = re.search('Found\s([0-9]+)\serrors', result)

    if fail and error:
        console.fail( ' :puke:\n' + error.group())


def jsdoc(files, folder, template = None, fail = True):
    System.check_package('java')

    files = FileList.check(files)

    if not template:
        template = "%s/templates/gris_taupe" % jsdoc

    jsdoc =  os.path.join(__get_datas_path(), 'jsdoc-toolkit')
    output = sh("java -jar %s/jsrun.jar %s/app/run.js -d=%s -t=%s -a  %s" % (jsdoc, jsdoc, folder, template, ' '.join(files)), header = "Generating js doc", output = True)

    if fail and output:
        console.fail(output)
    
    console.confirm('  Doc generated in "%s"' % folder)

def patch(dir, patch, p=0):
    # patch -p0 -N <

    output = sh(['cd %s' % dir, 'patch -p%s -N -r pukepatch.rej < %s' % (p, patch)], output = False)
    console.header(' - Patching %s with %s' % (dir, patch))
    lines = output.split('\n')
    for line in lines:
        if 'can\'t find file to patch' in line:
            console.fail('Path : can\'t find file to patch')
            return False
        
        if line.startswith('patching'):
            console.confirm('    ' + line)
        elif 'ignored' in line:
            console.warn('    ' + line)
        else:
            console.info('    ' + line)
    
    try:
        remove(join(dir, 'pukepatch.rej'))
    except:
        pass


def prompt(message, default = ''):
    console.warn(message)
    result = sh('read toto && echo $toto', header = None, output=False)
    result = result.strip()
    if not result:
        result = default
    
    return result


def deepcopy(file_list, folder, replace = None):

    file_list = FileList.check(file_list, True)

    stat = 0
    console.header( "- copy to %s (%s files)" % (folder, len(file_list)))


    #Check if Sed changed
    forceRefresh = False
    if replace:
        filesId = FileList.getSignature(file_list)
        sedID = replace.getSignature()
        filesInfo = Cache.read("sed-%s" % filesId)

        sedUpdated = True

        if filesInfo:
            lastSed = filesInfo.split('\n')[-1]
            lastSed = lastSed.split(':')[0]

            if lastSed == sedID:
                sedUpdated = False
            
            filesInfo += "\n"
        else:
            filesInfo = ""



        if sedUpdated:
            forceRefresh = True
            Cache.write("sed-%s" % (filesId), "%s%s:%s" % (filesInfo, sedID, int(time.time())))

    
        
    
    for (file, basepath) in file_list:
        
        if basepath:
            dst_file = __pretty(file).replace(basepath, '').strip(os.sep)
            dst_file = os.path.join(folder,dst_file)
        else:
            dst_file = os.path.join(folder,os.path.basename(__pretty(file)))

        #console.warn("File : " + file +  " dest " + dst_file)

        if not forceRefresh:
            res = copyfile(file, dst_file, force = True)
        else:
            copyfile(file, dst_file)
            res = True
        

        if res and replace:

            fh = open(dst_file)
            data = fh.read()
        
            data = __replace(data, replace)

            fh.close()
            writefile(dst_file, data)
            


        if res:
            console.info(' + %s' % __pretty(file))
            stat += 1
    
    console.confirm( "  %s files updated" % (stat))

    


def stats(file_list, title = ""):
    file_list = FileList.check(file_list)

    if not title:
        title = "Stats on files"

    console.header(" - %s :" % title)

    if len(file_list) == 0:
        console.info(  "   No files ")
        return False

    size = __exec("du -k  %s | cut  -f1 |(tr '\n' '+'; echo 0) | bc" % ' '.join(file_list))
    
    try:
        size = int(size)
    except Exception:
        return 0
    
    size = size * 1024

    lines = __exec("wc -l %s | tail -1" % ' '.join(file_list))
    lines =  re.findall(r'\d+(?:\.\d+)?', lines)
    if len(lines):
        try:
            lines = int(lines.pop(0))
        except:
            lines = 0
    else:
        lines = 0
    
    
    console.info(  "   ~ Files : %s" % len(file_list))
    console.info(  "   ~ Lines : %s  (%s per file)" % (lines, (lines / len(file_list))))
    console.info(  "   ~ Size : %s (%s per file)" % (hsizeof(size), hsizeof((size / len(file_list)))))
    console.info("")

    return (len(file_list), lines, size)



def pack (file_list, output):

    file_list = FileList.check(file_list)

    console.header( "- Packing files to %s (%s files)" % (output, len(file_list)))

    comp = Compress.open(output, "w")
     
    for file in file_list:
        console.info(' * %s' % __pretty(file))
        comp.add(file)

    comp.close()

    console.confirm(" %s packed" % output)

def unpack (pack_file, output, extract = None, verbose=True):
    
    console.header( "- Unpacking %s to %s " % (pack_file, output))

    if not Compress.check(pack_file):
        console.error(" %s is not a valid pack" % output)
        raise PukeError('%s is not a valid pack' % output)
        return

    comp = Compress.open(pack_file, "r")
    
    

    count = 0
    for fname in comp:
        if extract and extract not in fname:
            continue

        (data, infos) = comp.extract(fname)
        
       
        #folder
        if not data:
            continue
        
        output_file = os.path.join(output, fname)


        if exists(output_file) and os.path.getmtime(output_file) == infos.mtime:
            continue
        
        if verbose == True:
            console.info(' + %s' % __pretty(output_file))
        
        writefile(output_file, data, mtime=infos.mtime)

        if infos.mode:
            chmod(output_file, infos.mode)

        count += 1
    
    comp.close()
    console.confirm(" %s unpacked in %s (%s files)" % (pack_file, output, count))

### WIP
"""
def __jasmine(files):

    System.check_package('java')

    files = [files]

    envjs =  os.path.join(__get_datas_path(), 'envjs')
    output = sh("cd %s && java -jar %s -opt -1 %s/envjs.bootstrap.js %s" % (envjs, JS_RUNNER, envjs, " ".join(files)) , header = None, output = False)

    lines = output.split('\n')

    hasFailed = False

    for line in lines:
        if "Envjs" in line:
            continue

        if "Loading" in line:
            console.info('  * ' + line)
        elif "FAILED" in line:
            hasFailed = True
            console.error('  ' + line + '  ')
        elif "Suite" in line:
            console.log(' • ' + line)
        elif "Spec" in line or "Error" in line:
            console.info('   ' + line)
        elif "Passed:" in line:
            console.log(' ' + '-'*40)
            console.log('  '+ line)
        elif "Failed:" in line:
            console.log('  ' + line)
        elif "Total" in line:
            console.log('  ' + line)
            console.log(' ' + '-'*40)
        else:
            console.info('  ' + line)
    
    if not hasFailed: 
        console.confirm('  Tests success')
    else:
        console.fail("  Tests failure")
"""
    
def __replace(data, replace):
    for k in replace.keys():
        data = re.sub(k, replace.get(k), data)

    return data


def __pretty(filename):
    if filename.startswith('.pukecache'):
        return Cache.getInfo(filename.split(os.sep).pop())
    
    return filename

def __minify_css(in_file, out_file, verbose):
    options = ['-o "%s"' % out_file,
               '--type css']

    if verbose:
        options.append('-v')

    os.system('java -jar "%s" %s %s' % (CSS_COMPRESSOR,
                                          ' '.join(options), in_file))

def __get_ext(filename):
    return filename.split('.')[-1]


def __parse_scss(payload):
    css = Scss()
    return css.compile(payload)

def __minify_js(in_file, out_file, verbose):
    options = ['--js %s' % in_file,
               '--js_output_file %s' % out_file,
               '--warning_level QUIET']

    os.system('java -jar "%s" %s' % (JS_COMPRESSOR,
                                          ' '.join(options)))

def __get_datas_path():
    return os.path.join(os.path.dirname( __file__ ), 'datas')


def __exec(command):
    result = ""
    p = os.popen(command)
    for line in p.readlines():
        result += line + '\n'
   
    p.close()

    return result

def hsizeof(num):
    try:
        num = int(num)
    except Exception:
        return 0
    
    for x in ['bytes','KB','MB','GB','TB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0




#!/usr/bin/env python
# -*- coding: utf8 -*-

import re, platform, os, pwd
from puke.ToolsExec import *
from distutils import version


try:
	LOGIN = pwd.getpwuid(os.getuid())[0]
except:
	LOGIN = None

_PLATFORM = OS = platform.system()
MACOS = "Darwin"
LINUX = "Linux"
WINDOWS = "Windows"

re_v = re.compile('((\d+)\.(\d+)(\.(\d+))?([ab](\d+))?)')
re_c = re.compile('(==|<=|>=|<|>)(.*)')

def check_package(name, compVersion = None, platform = "all"):

	if platform.lower() == LINUX.lower() and _PLATFORM != LINUX:
		return True
	
	if MACOS.lower() in platform.lower() and _PLATFORM != MACOS:
		return True

	check = sh("which %s" % name, header = None, output = False)
	check = check.strip()
	
	if not check:
		
		if _PLATFORM == MACOS:

			check = sh("brew info %s | grep -i -E \"(not installed|error)\"" % name, header=None, output=False)
			check = check.lower().strip()

			if "not installed" in check or 'error' in check:
				console.error('%s is M.I.A' % name)
				console.log(' => "brew install %s"' % name)
				console.error('')
				console.fail('%s not installed' % name)
		else:
			check = sh("aptitude show %s | grep -i -E \"(unable to locate|not installed)\"" % name, header=None, output=False)
			check = check.lower().strip()
			
			if "unable to locate" in check or 'not installed' in check:
				console.error('%s is M.I.A' % name)
				console.log(' => "aptitude install %s"' % name)
				console.error('')
				console.fail('%s not installed' % name)
		
		
	

	pVersion = get_package_version(name)

	goodVersion = True

	if compVersion:
		(op, wishVersion) = re_c.match(compVersion).groups()

		if op == "==":
			goodVersion = (version.StrictVersion(pVersion) == version.StrictVersion(wishVersion))
		elif op == "<=":
			goodVersion = (version.StrictVersion(pVersion) <= version.StrictVersion(wishVersion))
		elif op == ">=":
			goodVersion = (version.StrictVersion(pVersion) >= version.StrictVersion(wishVersion))
		elif op == "<":
			goodVersion = (version.StrictVersion(pVersion) < version.StrictVersion(wishVersion))
		elif op == ">":
			goodVersion = (version.StrictVersion(pVersion) > version.StrictVersion(wishVersion))
		else:
			console.fail('bad comp %s' % op)

	if goodVersion == True:
		console.confirm('%s : OK (%s)' % (name,pVersion))
	else:
		console.error('%s : OK (%s not %s)' % (name,pVersion, compVersion))

		console.fail('Failed on version comp %s ' % name)

	return True


def get_package_version(name, index = 0):
	check = ['--version', '-v', 'hard']

	if check[index] == 'hard':
		if _PLATFORM == MACOS:

			version = sh("toto='%s';test=`brew info $toto | grep -Ei '(?not installed|error)'`;\
							 if [[ -n \"$test\" ]]; then echo \"\"; else echo `brew info $toto | head -n 1`; fi" % (name), header = None, output = False)
		else:
			version = sh("toto='%s';test=`aptitude show $toto | grep -Ei '(?not installed|unable to locate)'`;\
							 if [[ -n \"$test\" ]]; then echo \"\"; else echo `aptitude show $toto | grep -i \"version:\"`; fi" % (name), header = None, output = False)
	else:
		version = sh('%s %s' % (name,check[index]), header = None, output = False, timeout=3)

	version = re_v.findall(version)
	if version:
		return version[0][0]
	elif len(check) <= index + 1:
		return "0.0"
	else:
		return get_package_version(name, index+1)



#!/usr/bin/env python
# -*- coding: utf8 -*-


from setuptools import setup, find_packages
import sys, os
import pkg_resources


major, minor = sys.version_info[:2]

if major < 2 and minor < 6:
    raise Exception("Puke requires Python 2.6")
import logging

setup(
    name = "puke",
    version = "1.5.1",
    packages = ['puke'],

    scripts = [
       'bin/puke',
       'bin/puke.js.compress',
       'bin/puke.css.compress'
    ],

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires = ['pyscss', 'closure_linter', 'colorama', 'pyyaml', 'paramiko'],
    dependency_links = ['http://closure-linter.googlecode.com/files/closure_linter-latest.tar.gz'],
  
    # metadata for upload to PyPI
    author = "Emmanuel Tabard",
    author_email = "manu@webitup.fr",
    description = "Puke is a straightforward build system",
    license = "http://www.gnu.org/copyleft/gpl.html",
    keywords = "build system python",
    url = 'http://github.com/webitup/puke',
    include_package_data = True
)


#!/usr/bin/env python
# -*- coding: utf8 -*-

import os, logging
import fnmatch, re
import hashlib



from puke.Console import *
from puke.FileSystem import *
from puke.Cache import *



class FileList:

	def __init__(self, dir, filter = "*", exclude = ""):
		self.__list = []
		self.__dir = dir
		self.__filter = ""


		if dir.startswith('http') or dir.startswith('ftp'):
			self.__list.append((dir,''))
			return


		for tmp_f in filter.split(','):
			tmp_f = tmp_f.strip(' ')
			self.__filter = "%s|(%s)" % (self.__filter, fnmatch.translate(tmp_f))
		
		self.__filter = self.__filter.strip('|')
		self.__filter = re.compile(self.__filter)
		
		self.__exclude = ""
		for tmp_f in exclude.split(','):
			tmp_f = tmp_f.strip(' ')
			self.__exclude = "%s|(%s)" % (self.__exclude, fnmatch.translate(tmp_f))
		
		self.__exclude = self.__exclude.strip('|')
		self.__exclude = re.compile(self.__exclude)



		self.__explore(dir)

	def merge (self, flist):
		if isinstance(flist, FileList):
			flist = flist.get(True)
		elif isinstance(flist, str):
			flist = [(flist, '')]
		else:
			result = []
			for f in flist:
				result.append( ( f, ''))
			flist = result
			del result

		for (f, p) in flist:
			if f in self.__list:
				continue
			
			self.__list.append( (f, p ) )
		
		
		
		return self.__list
			
	def get (self, full = False):
		if not full:
			result = []
			for (f, p) in self.__list:
				result.append(f)
			
			return result

		return self.__list

	
	@staticmethod
	def getSignature(list):
		sig = ""
		for (f,p) in list:
			sig += "%s" % hashlib.sha256(f).hexdigest()
		
		sig = hashlib.sha256(sig).hexdigest()

		return sig

	@staticmethod
	def check(flist, full = False):
		if isinstance(flist, FileList):
			flist = flist.get(True)
		elif isinstance(flist, str):

			try:
				cIsFile = isfile(flist)
			except:
				cIsFile = False
				
			flist = [(flist, '')] if cIsFile else FileList(flist).get(True)
		else:
			result = []
			for f in flist:
				result.append( ( f, ''))
			flist = result
			del result


		result = []
		for (f, p) in flist:
			if f.startswith('http') or f.startswith('ftp'):

				f = Cache.fetchHttp(f)
			
			if not full:
				result.append(f)
			else:
				result.append((f, p))
      
		return result

	def __explore(self, dir ):
	    dir = os.path.abspath(dir)
	    for file in sorted([file for file in os.listdir(dir) if not file in [".",".."]]):
	        nfile = os.path.join(dir,file)
	        
	        if os.path.isdir(nfile):
	            self.__explore(nfile)
	        elif self.__filter.match(nfile) and not self.__exclude.match(nfile):
	        	tmp = nfile.split(self.__dir)
	        	if len(tmp) > 1:
	        		nfile = self.__dir + tmp[1]
	        	else:
	        		nfile = tmp[0]
	        	
	        	
	        	self.__list.append( (nfile, self.__dir))





		       



class Std:

	def __init__(self, out = '', err = '', code = 0):
		self.out = out
		self.err = err
		self.code = code

	
	def set(self, out, err, code = 0):
		self.out = out
		self.err = err
		self.code = code

	def __repr__(self):
		return "Stdout : %s\n\nStdErr : %s\n" % (self.out, self.err, self.code)

	def __str__(self):
		return self.__repr__()


#!/usr/bin/env python
# -*- coding: utf8 -*-

from puke.Tools import *
from puke.ToolsExec import *
from puke.FileSystem import *
import System
from distutils import version

import os


re_v = re.compile('((\d+)\.(\d+)(\.(\d+))?([ab](\d+))?)')
re_c = re.compile('(==|<=|>=|<|>)?(.*)')

class VirtualEnv(object):
	def __init__(self):
		self.__path = None

	def load (self, path):
		self.__path = path

	def create(self, path, python = "python", force = False):

		console.header(' * Creating env "%s" ...' % (path))
		if not force and exists(os.path.join(path, 'bin', 'activate')):
			self.__path = path
			console.confirm('   Env "%s" already created. Use "force=True" to override it' % path)
			return True

		check = System.check_package('virtualenv')
		version = System.get_package_version(python)
		console.log('   Python version : %s ...' % version)

		

		result = sh("virtualenv -p %s %s --no-site-packages " % (python, path), header = False, output = False)
		console.debug(result)

		console.confirm('   Env "%s"  created' % path)
		self.__path = path


	def install(self, package, whishVersion = None, upgrade = False):
		if not self.__path:
			raise VirtualEnvError('Need load|create before install')

		if not upgrade:
			console.header(' * Install "%s" in env "%s" ...' % (package, self.__path))
		else:
			console.header(' * Upgrade "%s" in env "%s" ...' % (package, self.__path))

		opts = []

		if upgrade == True:
			opts.append('--upgrade')

		cmd = [
			'source %s' % os.path.join(self.__path, 'bin', 'activate'),
			'pip install %s %s' % (package if not whishVersion else package+"=="+whishVersion, " ".join(opts))
		]
		result = sh(cmd, header = False, output = False)
		console.debug(result)

		console.confirm('   Package "%s" is ready' % package)

	def upgrade(self, package = "*"):

		if package == "*":
			packages = self.list(output = False)
		else:
			packages = [(package, 0)]

		for (package, v) in packages:
			self.install(package, True)

	def check_package(self, package, compVersion = None, fix = False):

		console.header('  * Check package "%s" %s ' % (package, compVersion))

		infos = self.package_info(package, output = False)

		op = wishVersion = None
		if compVersion:
			
			if compVersion.lower() != "latest":
				(op, wishVersion) = re_c.match(compVersion).groups()

			compVersion = compVersion.lower()

		if not infos:
			console.error('    No package')

			if fix:
				console.log('   Fixing ...')
				self.install(package, wishVersion)
				return True
			else:
				return False
		else:
			(name, cVersion, lVersion) = infos


		if compVersion and compVersion != "latest":
			goodVersion = False
			if op == "==":
				goodVersion = (version.StrictVersion(cVersion) == version.StrictVersion(wishVersion))
			elif op == "<=":
				goodVersion = (version.StrictVersion(cVersion) <= version.StrictVersion(wishVersion))
			elif op == ">=":
				goodVersion = (version.StrictVersion(cVersion) >= version.StrictVersion(wishVersion))
			elif op == "<":
				goodVersion = (version.StrictVersion(cVersion) < version.StrictVersion(wishVersion))
			elif op == ">":
				goodVersion = (version.StrictVersion(cVersion) > version.StrictVersion(wishVersion))
			else:
				console.fail('   bad comparator %s (eg : >=1.6)' % op)			

			if not goodVersion:

				if op == ">":
					wishVersion = lVersion

				console.error('    Wrong version %s VS %s' % (cVersion, wishVersion))

				if fix:
					console.log('   Fixing ...')
					return self.install(package, wishVersion, upgrade = True)
				else:
					return False

			else:
				console.confirm(' Everything is ok (Current %s, Last %s)' % (cVersion, lVersion))
				return True

		elif compVersion == "latest":
			if (version.StrictVersion(cVersion) == version.StrictVersion(lVersion)):
				console.confirm('    This is the latest package')
				return True
			else:
				console.warn('    Package needs an upgrade %s => %s' % (cVersion, lVersion))
				if fix:
					console.log('   Fixing ...')
					return self.install(package, wishVersion, upgrade = True)
				else:
					return False
		else:
			console.confirm('    Everything is ok (Current %s, Last %s)' % (cVersion, lVersion))

		return True

		


	def package_info(self, package, output = True):

		if output:
			console.header('  * Package "%s" :' % package)
		
		package = package.lower().strip()
		installed_packages = self.list(output = False)

		here = False
		info = {}
		for (p, v) in installed_packages:
			if package != p.lower().strip():
				continue
			else:
				here = True
				info['name'] = p
				info['version'] = v
				break
		
		if not here:
			if output:
				console.error('    No package "%s"' % package)
			return False
		
		#get last version
		cmd = [
			'source %s' % os.path.join(self.__path, 'bin', 'activate'),
			'pip search %s | grep "LATEST:" | head -n 1' % package
		]
		result = sh(cmd, header = False, output = False)

		last = re_v.findall(result)
		if last:
			last = last[0][0]
		else:
			last = info['version']

		if output:
			

			if last != "0.0" and last != info['version']:
				console.confirm('    - Current Version %s' % info['version'])
				console.warn('    - Latest Version %s ' % last)
			else:
				console.confirm('    - Current Version %s (latest)' % info['version'])

		return (info['name'], info['version'], last)


	def list(self, output = True):
		if not self.__path:
			raise VirtualEnvError('Need load|create before install')
		
		cmd = [
			'source %s' % os.path.join(self.__path, 'bin', 'activate'),
			'pip freeze'
		]

		if output:
			console.header(' * List packages (env \"%s\")' % self.__path)


		result = sh(cmd, header = False, output = False)
		
		lines = result.split('\n')


		packages = []
		for line in lines:
			if line == "":
				continue

			extract = line.rsplit('==', 1)
			if len(extract) == 1:
				name = extract[0]
				version = 0
			else:
				(name, version) = extract

			if output:
				console.info('   - %s (%s)' % (name, version))

			packages.append((name, version))
		
		return packages
	
	def remove(self):
		 if not self.__path or not exists(os.path.join(self.__path, 'bin', 'activate')):
		 	raise VirtualEnvError('"%s" is not a python env')
		 
		 console.header(' * Deleting env \"%s\" ... ' % self.__path)
		 rm(self.__path)
		 console.confirm('   Env "%s" deleted' % self.__path)



class VirtualEnvError(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)
